#include "PreCompiledHeaders.hh"
#include "Lib/Ass/AssFactory.hh"

using namespace Vivy::Ass;

bool
AssFactory::initFromStorage() noexcept
{
    QTextStream in(&diskStorage);
    QString currentSection{};
    quint64 lineIndex = 0;
    QStringList stylesContent{};
    QStringList eventsContent{};

    while (!in.atEnd()) {
        const QString line = in.readLine().trimmed();

        // Dectect comment
        if (line.startsWith(";") || line.isEmpty()) {
            lineIndex++;
            continue;
        }

        // Dectect sections
        else if (line.startsWith("[") && line.endsWith("]")) {
            currentSection = line.mid(1, line.size() - 2);
            logDebug() << "Parsing section " << VIVY_LOG_QUOTED(currentSection);
            if (!validSections.contains(currentSection)) {
                logWarning() << "The current section " << VIVY_LOG_QUOTED(currentSection)
                             << " is invalid, ignoring it";
                currentSection = "";
            }
        }

        // Other lines
        else if (!currentSection.isEmpty()) {
            const int separatorIndex = line.indexOf(": ");
            const int baseValueIndex = separatorIndex + 2;

            // Easy way to see if the line was invalid
            if (separatorIndex < 0)
                logWarning() << "Invalid line #" << lineIndex << ": " << line;

            // Script's info
            else if (currentSection == sectionScriptInfo) {
                assInfo.insert(line.mid(0, separatorIndex), line.mid(baseValueIndex));
                logDebug() << "Got line #" << lineIndex << ": " << line;
            }

            // Skip the headers and the comment lines
            else if ((currentSection == sectionStyles) && line.startsWith("Style: ")) {
                stylesContent.append(line);
            } else if ((currentSection == sectionEvents) && line.startsWith("Dialogue: ")) {
                eventsContent.append(line);
            } else if ((currentSection == sectionEvents) && line.startsWith("Comment: ")) {
                eventsContent.append(line);
            }
        }

        lineIndex++;
    }

    // Construct the styles and the lines
    try {
        for (const auto &styleLine : stylesContent)
            assStyles.push_back(std::make_shared<Style>(styleLine));
        for (const auto &assLine : eventsContent)
            assLines.push_back(std::make_shared<Line>(this, assLine));
    } catch (const std::runtime_error &e) {
        logError() << "Failed to create ASS style or events with error: " << e.what();
    }

    logDebug() << "Got " << assLines.size() << " ASS dialog lines";

    return true;
}

bool
AssFactory::checkValidity() const noexcept
{
    if (assInfo.isEmpty()) {
        logError() << "Empty info section";
        return false;
    }

    // Check for fields that must be integers
    SectionContent::const_iterator it        = assInfo.begin();
    const SectionContent::const_iterator end = assInfo.end();
    while (it != end) {
        bool ok = false;
        if (intTypeFields.contains(it.key()) && (static_cast<void>(it.value().toInt(&ok)), !ok)) {
            logError() << it.key() << " is not an integer: " << it.value();
            return false;
        }
        ++it;
    }

    // Check for fixed values fields
    for (const auto &fixedValues : checkedValues) {
        if (const auto &first = fixedValues.first;
            assInfo.contains(first) && assInfo[first] != fixedValues.second) {
            logError() << "Invalid " << first << " as it should be equal to " << fixedValues.second
                       << " but was " << assInfo[first];
            return false;
        }
    }

    return true;
}

AssFactory::AssFactory(const QString &fileName)
    : diskStorage(fileName)
{
    if (!diskStorage.open(QIODevice::ReadOnly | QIODevice::Text))
        throw std::runtime_error("failed to open file for reading");

    if (!initFromStorage())
        throw std::runtime_error("failed to init ass factory from file");

    if (!checkValidity())
        throw std::runtime_error("the loaded ass is invalid");
}

void
AssFactory::getStyles(QVector<StylePtr> &ret) const noexcept
{
    ret.clear();
    for (const auto &style : assStyles)
        ret.push_back(style);
}

void
AssFactory::getLines(QVector<LinePtr> &ret) const noexcept
{
    ret.clear();
    for (const auto &line : assLines)
        ret.push_back(line);
}

AssFactory::SectionContent
AssFactory::getInfoSection() const noexcept
{
    return assInfo;
}

StyleWeakPtr
AssFactory::getStyle(const QString &name) const noexcept
{
    auto findByName = [&name](const StylePtr &style) noexcept -> bool {
        return style->getElementName() == name;
    };

    auto styleIt = std::find_if(std::begin(assStyles), std::end(assStyles), findByName);
    if (styleIt != std::end(assStyles))
        return StyleWeakPtr(*styleIt);

    // Will be unable to lock
    return StyleWeakPtr(spareNullStylePtr);
}

bool
AssFactory::hasStyle(const QString &name) const noexcept
{
    for (const auto &stylePtr : assStyles) {
        if (stylePtr->getElementName() == name)
            return true;
    }
    return false;
}
