#pragma once

#ifndef __cplusplus
#error "This is a C++ header"
#endif

#include "PreCompiledHeaders.hh"
#include "Lib/AbstractMediaContext.hh"

namespace Vivy
{
class AudioContext;

// Hold all the data for a stream in an audio file. Should only be owned by
// the AudioContext class.
class AudioStream final : public AbstractMediaStream<AVMEDIA_TYPE_AUDIO> {
    VIVY_UNMOVABLE_OBJECT(AudioStream)

public:
    AudioStream(const AVCodec *, AVFormatContext *, AVStream *, int index);
    ~AudioStream() noexcept override;

    // Decode the stream
    void decodeData();
    void cleanUpData() noexcept;
    bool isDecoded() const noexcept;
    int getDataSampleRate() const noexcept;
    size_t getDecodedDataSize() const noexcept;
    size_t getDecodedChunkSize() const noexcept;
    size_t getDecodedDecalage() const noexcept;
    double *getDecodedData() const noexcept;

    // Some getters
    int getChannels() const noexcept;
    int getSampleRate() const noexcept;
    qint64 getBitRate() const noexcept;
    QJsonObject getProperties() const noexcept override;

private:
    // Resampled frame data
    SwrContextPtr dataSwrContext{ swr_alloc(), swrContenxtDeleter };
    AVFramePtr dataFrame{ av_frame_alloc(), avFrameDeleter };
    double *dataPtr{ nullptr };
    size_t dataSize{ 0 };

    // Constants for the resampler
    static constexpr uint resamplerChunkSize  = 512;
    static constexpr uint resamplerOverlap    = 128;
    static constexpr uint resamplerDecalage   = resamplerChunkSize - resamplerOverlap;
    static constexpr uint resamplerSampleRate = 44100;
};

// The memory representation of an audio file. It is created inplace and should
// not be moved around. It should be used to get the properties of an audio
// file and its streams. The user can use a stream from the file and get data
// from it.
class AudioContext final : public AbstractMediaContext<AudioStream> {
    VIVY_UNMOVABLE_OBJECT(AudioContext)

public:
    using StreamPtr     = std::shared_ptr<AudioStream>;
    using StreamWeakPtr = std::weak_ptr<AudioStream>;

public:
    AudioContext(const QString &path);

    QString getElementName() const noexcept override;
    QJsonDocument getProperties() const noexcept override;
};
}
