#pragma once

#include "PreCompiledHeaders.hh"
#include "VivyApplication.hh"
#include "Lib/Log.hh"
#include "Lib/Script/LuaContext.hh"
#include "Lib/Script/ScriptOption.hh"

#include "lua.hpp"

namespace Vivy::Script
{
// clang-format off
#define LUA_DECL_METHOD(class, name)                luaL_Reg{ #name,  class ::name   }
#define LUA_DECL_NAMED_METHOD(class, method, name)  luaL_Reg{ method, class ::name   }
#define LUA_DECL_META_METHOD(class, meta, name)     luaL_Reg{ meta,   class ::name   }
#define LUA_DECL_CREATE(class)                      luaL_Reg{ "new",  class ::CREATE }
#define LUA_RETURN_NOTHING(L)                       { lua_settop(L, 0); return 0; }
// clang-format on

#define method_list static constexpr inline auto
#define script_class(theClassName) \
    class theClassName final : public CRTPLuaScriptObject<theClassName>

#define LUA_DECL_GETTER(type, luaName, value) \
    static int luaName(lua_State *const L) noexcept { return luaPushValue(L, CHECK(L, 1)->value); }

#define LUA_SCRIPTABLE_CLASS_MANUAL_CTOR(theClassName) \
    VIVY_UNMOVABLE_OBJECT(theClassName)                \
    static constexpr char className[] = #theClassName; \
    friend CRTPLuaScriptObject<theClassName>;          \
    ~theClassName() noexcept {}

#define LUA_SCRIPTABLE_CLASS(theClassName)                               \
    LUA_SCRIPTABLE_CLASS_MANUAL_CTOR(theClassName)                       \
    theClassName(lua_State *L) noexcept                                  \
    {                                                                    \
        err(LuaContext::getContext(L))                                   \
            << "Create instance of " << theClassName::className << "\n"; \
    }

// The type of the thing that a job iterate over
enum class JobIteratorType { Line = 1, Syllabe = 2 };

[[maybe_unused]] static inline JobIteratorType
getJobIteratorTypeFromString(const std::string_view it) noexcept
{
    if (it == "LINE")
        return JobIteratorType::Line;
    else if (it == "SYLLABE")
        return JobIteratorType::Syllabe;
    return static_cast<JobIteratorType>(0); // Invalid!
}

// CRTP to expose objects to Lua
template <class Object> class CRTPLuaScriptObject {
protected:
    VIVY_APP_LOGGABLE_OBJECT_BY_STORED_NAME(Object::className, logger)

    [[noreturn]] static void luaGlobalError(lua_State *const L, const std::string &str) noexcept
    {
        const auto *const context = LuaContext::getContext(L);
        lua_pushfstring(L, "%s:0:%s", context->getCurrentLuaFile().c_str(), str.c_str());
        lua_error(L);
        exit(EXIT_FAILURE); // lua_error should not return anything
    }

    // Push something onto the stack, must be specialized
    using std_str  = std::string;
    using std_strv = std::string_view;

    static int luaPushValue(lua_State *const L, const int &thing) noexcept
    {
        lua_pushinteger(L, thing);
        return 1;
    }

    static int luaPushValue(lua_State *const L, const std_strv &thing) noexcept
    {
        lua_pushlstring(L, thing.data(), thing.size());
        return 1;
    }

    static int luaPushValue(lua_State *const L, const std_str &thing) noexcept
    {
        lua_pushstring(L, thing.c_str());
        return 1;
    }

    static bool PushFunctionFromRegistry(lua_State *const L, const int key) noexcept
    {
        if (key >= 0) {
            lua_rawgeti(L, LUA_REGISTRYINDEX, key);
            return true;
        } else {
            lua_pushnil(L);
            return false;
        }
    }

    static void IterateOverArray(lua_State *const L, const int tblIndex,
                                 const auto callback) noexcept
    {
        luaL_checktype(L, tblIndex, LUA_TTABLE);
        const std::size_t count = lua_rawlen(L, tblIndex);
        for (int i = 1; std::cmp_less(i, count); ++i) {
            luaL_checktype(L, tblIndex, LUA_TTABLE);
            lua_pushinteger(L, i);     // i -> top stack
            lua_gettable(L, tblIndex); // T[i] -> top stack and pop i

            if (lua_type(L, -1) == LUA_TNIL)
                break; // Sentinel check

            // If the callback didn't pop the object form the stack, pop
            // it ourself. The value will be accessible at the top of the
            // stack (index -1).
            if (!callback())
                lua_pop(L, -1);
        }
    }

    static int GC(lua_State *const L) noexcept
    {
        const Object *const obj = reinterpret_cast<const Object *>(lua_topointer(L, 1));
        auto *context           = LuaContext::getContext(L);
        out(context) << "Call destructor for " << Object::className;
        obj->~Object();
        return 0;
    }

    static int TOSTRING(lua_State *const L) noexcept
    {
        lua_pushfstring(L, "%s { }", Object::className);
        return 1;
    }

    static int CREATE(lua_State *const L) noexcept
    {
        lua_settop(L, 0);
        Object *const obj = reinterpret_cast<Object *>(lua_newuserdata(L, sizeof(Object)));
        new (obj) Object(L);
        luaL_getmetatable(L, Object::className);
        lua_setmetatable(L, 1);
        return 1;
    }

    static const std::string_view CHECK_STRING_VIEW(lua_State *const L, int index) noexcept
    {
        std::size_t len       = 0;
        const char *const str = luaL_checklstring(L, index, &len);
        return std::string_view(str, len);
    }

    static inline constexpr luaL_Reg luaRegDefaultGC       = { "__gc", GC };
    static inline constexpr luaL_Reg luaRegDefaultTOSTRING = { "__tostring", TOSTRING };

public:
    static Object *CHECK(lua_State *const L, const int index) noexcept
    {
        return reinterpret_cast<Object *>(luaL_checkudata(L, index, Object::className));
    }

    static void Register(lua_State *const L) noexcept
    {
        // Fill the method table, newclass = {}
        lua_newtable(L);
        const int methodtable = lua_gettop(L);

        // Fill the meta-table, metatable = {}
        {
            luaL_newmetatable(L, Object::className);
            const int metatable = lua_gettop(L);

            lua_pushliteral(L, "__metatable");
            lua_pushvalue(L, methodtable);
            lua_settable(L, metatable);

            lua_pushliteral(L, "__index");
            lua_pushvalue(L, methodtable);
            lua_settable(L, metatable);

            for (const luaL_Reg &meta : Object::metaMethods) {
                lua_pushstring(L, meta.name);
                lua_pushcfunction(L, meta.func);
                lua_settable(L, metatable);
            }

            lua_pop(L, 1);
        }

        // Fill the method table with specified methods
        for (const luaL_Reg &method : Object::methods) {
            lua_pushstring(L, method.name);
            lua_pushcfunction(L, method.func);
            lua_settable(L, methodtable);
        }

        // Last thing to do it seems...
        lua_pop(L, 1);                              // drop method-table
        lua_register(L, Object::className, CREATE); // Register the create method
    }

private:
    // The user value count
    int userValueCount{ 0 };

protected:
    // Add a user value to this object. Do not used lua_setuservalue or
    // lua_setiuservalue directly, use this proxy method.
    void addUserValue(lua_State *const L, const int index) noexcept
    {
        lua_setiuservalue(L, index, ++userValueCount);
    }
};

// Option declaration
script_class (OptionDeclaration) {
    LUA_SCRIPTABLE_CLASS(OptionDeclaration)

    static int addOptionContent(lua_State *const) noexcept;

    static int TOSTRING(lua_State *const) noexcept;

    method_list metaMethods = {
        luaRegDefaultGC,
        LUA_DECL_META_METHOD(OptionDeclaration, "__tostring", TOSTRING),
    };
    method_list methods = { LUA_DECL_METHOD(OptionDeclaration, addOptionContent),
                            LUA_DECL_CREATE(OptionDeclaration) };

public:
    // Different option types
    std::unordered_map<std::string, ScriptOption> options;
};

// Job declaration
script_class (JobDeclaration) {
    LUA_SCRIPTABLE_CLASS(JobDeclaration)

    static int setName(lua_State *) noexcept;
    static int setOptions(lua_State *) noexcept;
    static int setImportFromScript(lua_State *) noexcept;

    static int pushToRuntime(lua_State *const) noexcept;

    static int TOSTRING(lua_State *const) noexcept;

    method_list metaMethods = { luaRegDefaultGC,
                                LUA_DECL_META_METHOD(JobDeclaration, "__tostring", TOSTRING) };
    method_list methods     = { LUA_DECL_METHOD(JobDeclaration, setName),
                            LUA_DECL_METHOD(JobDeclaration, setOptions),
                            LUA_DECL_METHOD(JobDeclaration, setImportFromScript),
                            LUA_DECL_METHOD(JobDeclaration, pushToRuntime),
                            LUA_DECL_CREATE(JobDeclaration) };

    bool isImported{ false };
    int functionRegisterKey{ -1 };
    JobIteratorType iterationType{ static_cast<JobIteratorType>(0) }; // Invalid by default

public:
    std::string name{};
    std::string parentScript{};
    std::vector<OptionDeclaration *> options{};

    bool isLocalToModule() const noexcept;
    void pushLuaFunction(lua_State *const) const noexcept;
    void getLuaFunctionFromStack(lua_State *const, const JobIteratorType) noexcept;
};

// Function declaration
script_class (FunctionDeclaration) {
    LUA_SCRIPTABLE_CLASS(FunctionDeclaration)

    static int setName(lua_State *const) noexcept;
    static int setImportFromScript(lua_State *const) noexcept;

    static int pushToRuntime(lua_State *const) noexcept;

    static int TOSTRING(lua_State *const) noexcept;

    method_list metaMethods = { luaRegDefaultGC,
                                LUA_DECL_META_METHOD(FunctionDeclaration, "__tostring", TOSTRING) };
    method_list methods     = { LUA_DECL_METHOD(FunctionDeclaration, setName),
                            LUA_DECL_METHOD(FunctionDeclaration, setImportFromScript),
                            LUA_DECL_METHOD(FunctionDeclaration, pushToRuntime),
                            LUA_DECL_CREATE(FunctionDeclaration) };

    bool isImported{ false };
    int functionRegisterKey{ -1 };

public:
    std::string name{};
    std::string parentScript{};

    bool isLocalToModule() const noexcept;
    void pushLuaFunction(lua_State *const) const noexcept;
    void getLuaFunctionFromStack(lua_State *const) noexcept;
};

// Module declaration
script_class (ModuleDeclaration) {
    LUA_SCRIPTABLE_CLASS(ModuleDeclaration)

    static int setName(lua_State *const) noexcept;
    static int setAuthorName(lua_State *const) noexcept;
    static int setRevision(lua_State *const) noexcept;
    static int setDescription(lua_State *const) noexcept;
    static int setImports(lua_State *const) noexcept;
    static int setOptions(lua_State *const) noexcept;
    static int setFunctions(lua_State *const) noexcept;
    static int setJobs(lua_State *const) noexcept;

    static int pushToRuntime(lua_State *const) noexcept;

    static int importMethod(lua_State *const) noexcept;
    static int exportMethod(lua_State *const) noexcept;

    method_list metaMethods = { luaRegDefaultGC };
    method_list methods     = {
        LUA_DECL_METHOD(ModuleDeclaration, setName),
        LUA_DECL_METHOD(ModuleDeclaration, setDescription),
        LUA_DECL_METHOD(ModuleDeclaration, setRevision),
        LUA_DECL_METHOD(ModuleDeclaration, setAuthorName),
        LUA_DECL_METHOD(ModuleDeclaration, setImports),
        LUA_DECL_METHOD(ModuleDeclaration, setOptions),
        LUA_DECL_METHOD(ModuleDeclaration, setFunctions),
        LUA_DECL_METHOD(ModuleDeclaration, setJobs),
        LUA_DECL_METHOD(ModuleDeclaration, pushToRuntime),
        LUA_DECL_NAMED_METHOD(ModuleDeclaration, "import", importMethod),
        LUA_DECL_NAMED_METHOD(ModuleDeclaration, "export", exportMethod),
        LUA_DECL_CREATE(ModuleDeclaration),
    };

    // Will be resolved later
    std::vector<std::string> importNames{};
    std::vector<const ModuleDeclaration *> importedModules{};

    void validateModule(lua_State *const) const noexcept;
    int exportFunction(lua_State *const, const int tableIndex, const std::string_view) noexcept;
    int exportJob(lua_State *const, const int tableIndex, const std::string_view,
                  const std::string_view) noexcept;

public:
    std::string moduleName{};
    std::string authorName{};
    std::string revision{};
    std::string moduleDescription{};
    std::vector<OptionDeclaration *> moduleOptions{};
    std::vector<FunctionDeclaration *> moduleFunctions{};
    std::vector<JobDeclaration *> moduleJobs{};

    bool resolvModules(LuaContext *const) noexcept;
};

// Holds all the free functions (well, not really free functions here...)
script_class (FreeFunctions) {
    LUA_SCRIPTABLE_CLASS_MANUAL_CTOR(FreeFunctions)

    FreeFunctions(lua_State *const) noexcept;

    static int print(lua_State *const) noexcept;
    static int getModule(lua_State *const) noexcept;

    LUA_DECL_GETTER(int, start, audioStart)
    LUA_DECL_GETTER(int, finish, audioFinish)
    LUA_DECL_GETTER(int, width, videoWidth)
    LUA_DECL_GETTER(int, height, videoHeight)

    method_list metaMethods = { luaRegDefaultGC };
    method_list methods     = { LUA_DECL_METHOD(FreeFunctions, getModule),
                            LUA_DECL_METHOD(FreeFunctions, start),
                            LUA_DECL_METHOD(FreeFunctions, finish),
                            LUA_DECL_METHOD(FreeFunctions, width),
                            LUA_DECL_METHOD(FreeFunctions, height),
                            LUA_DECL_METHOD(FreeFunctions, print),
                            LUA_DECL_CREATE(FreeFunctions) };

    int videoWidth{ 0 }, videoHeight{ 0 };
    int audioStart{ 0 }, audioFinish{ 0 };
};
}
