#include "PreCompiledHeaders.hh"
#include "Lib/Script/CRTPLuaScriptObject.hh"
#include "Lib/Script/ScriptDocument.hh"
#include "Lib/Script/LuaContext.hh"
#include "Lib/Document/VivyDocument.hh"

// LuaContext implementation
using namespace Vivy::Script;

LuaContext::LuaContext() noexcept
    : L(luaL_newstate())
{
    luaL_openlibs(L);

    // Register this context
    contextList.emplace(L, this);

    // Load all C++ objects
    FreeFunctions::Register(L);
    OptionDeclaration::Register(L);
    JobDeclaration::Register(L);
    FunctionDeclaration::Register(L);
    ModuleDeclaration::Register(L);

    // Load the lib
    loadPackagedFile(":lua/lib.lua");
    loadPackagedFile(":lua/module.lua");
}

LuaContext::~LuaContext() noexcept
{
    lua_close(L);         // Close lua
    contextList.erase(L); // Unregister the context
}

void
LuaContext::loadPackagedFile(const QString &url) noexcept
{
    currentFile = url.toStdString();
    QFile libVivy(url);
    if (!libVivy.open(QIODevice::ReadOnly | QIODevice::Text))
        qFatal("FATA -> failed to find packaged Vivy file %s", url.toStdString().c_str());
    loadString(libVivy.readAll().toStdString().c_str());
    currentFile = "";
}

LuaContext *
LuaContext::getContext(const lua_State *const state) noexcept
{
    return contextList.contains(state) ? contextList.at(state) : nullptr;
}

LuaContext::Code
LuaContext::loadDocument(std::shared_ptr<Vivy::VivyDocument> doc) noexcept
{
    if (attachedVivyDocument != nullptr)
        return Error;
    attachedVivyDocument = doc;
    return Success;
}

LuaContext::Code
LuaContext::loadDocument(std::shared_ptr<ScriptDocument> doc) noexcept
{
    const QFileInfo fileCheck(doc->getName());
    if (!fileCheck.exists()) {
        err(this) << "File " << doc->getName().toStdString() << " for document "
                  << doc->getUuid().toString().toStdString() << " doesn't exists";
        return Code::Error;
    }

    lastLoadedModule = "";
    const auto rc    = loadFile(fileCheck.absoluteFilePath().toStdString().c_str());
    if (rc != Code::Success) {
        setFailedWith("Failed to load the module file");
        return rc;
    }

    if (lastLoadedModule.empty()) {
        setFailedWith("No module was loaded");
        return Code::Error;
    }

    return Code::Success;
}

void
LuaContext::swapEnv() noexcept
{
    lua_setglobal(L, "vivy");
}

LuaContext::Code
LuaContext::exec() noexcept
{
    if (lua_pcall(L, 0, 0, 0) != LUA_OK) {
        err(this) << getLastLuaError().toStdString().c_str();
        return Code::Error;
    }
    return Code::Success;
}

LuaContext::Code
LuaContext::loadString(const char *str) noexcept
{
    if (luaL_loadstring(L, str) != LUA_OK) {
        err(this) << "Error loading string: " << getLastLuaError();
        return Code::Error;
    }
    return exec();
}

LuaContext::Code
LuaContext::loadFile(const char *file) noexcept
{
    currentFile = file;
    if (luaL_loadfile(L, file) != LUA_OK) {
        err(this) << "Error loading file " << file << ": " << getLastLuaError();
        currentFile = "";
        return Code::Error;
    }
    auto rc     = exec();
    currentFile = "";
    return rc;
}

QString
LuaContext::getLastLuaError() noexcept
{
    const int type = lua_type(L, -1);
    if (type != LUA_TSTRING) {
        return QStringLiteral("No lua error detected...");
    }

    QString ret = QString::fromUtf8(lua_tostring(L, -1));
    lua_pop(L, 1);
    return ret;
}

void
LuaContext::failedWith(const std::string &errorMessage) noexcept
{
    failedWith(errorMessage.c_str());
}

void
LuaContext::failedWith(const char *errorMessage) noexcept
{
    setFailedWith(errorMessage);
    lua_error(L);
    exit(EXIT_FAILURE); // Should not be there!
}

void
LuaContext::setFailedWith(const char *errorMessage) noexcept
{
    lua_pushstring(L, errorMessage);
}

void
LuaContext::setFailedWith(const std::string &errorMessage) noexcept
{
    setFailedWith(errorMessage.c_str());
}

lua_State *
LuaContext::getState() noexcept
{
    return L;
}

std::shared_ptr<Vivy::VivyDocument>
LuaContext::getAttachedVivyDocument() const noexcept
{
    return attachedVivyDocument;
}

bool
LuaContext::moduleExists(const std::string_view mod) const noexcept
{
    return modules.count(mod) >= 1;
}

LuaContext::Code
LuaContext::registerDeclaration(const FunctionDeclaration *const fun) noexcept
{
    FunctionDeclaration::CHECK(L, -1);
    return registerToMapTuple(fun->parentScript, fun->name, &functions);
}

LuaContext::Code
LuaContext::registerDeclaration(const JobDeclaration *const job) noexcept
{
    JobDeclaration::CHECK(L, -1);
    return registerToMapTuple(job->parentScript, job->name, &jobs);
}

LuaContext::Code
LuaContext::registerToMapTuple(const std::string_view module, const std::string_view name,
                               std::map<std::tuple<const std::string_view, const std::string_view>,
                                        const int> *const map) noexcept
{
    if (map->count({ module, name }) >= 1)
        return Error;

    logInfo() << "Register " << module << "::" << name;

    // WARN: The object needs to be already checked!
    const int r = luaL_ref(L, LUA_REGISTRYINDEX);
    map->emplace(std::tuple{ module, name }, r);
    return Success;
}

LuaContext::Code
LuaContext::registerDeclaration(const ModuleDeclaration *const mod) noexcept
{
    if (moduleExists(mod->moduleName)) {
        logError()
            << "The module " << VIVY_LOG_QUOTED(mod->moduleName) << " is already registered!";
        return Code::Error;
    }

    logInfo() << "Register module " << VIVY_LOG_QUOTED(mod->moduleName);

    ModuleDeclaration::CHECK(L, -1);
    const int r = luaL_ref(L, LUA_REGISTRYINDEX);
    modules.emplace(mod->moduleName, r);
    lastLoadedModule = mod->moduleName;
    return Code::Success;
}

ModuleDeclaration *
LuaContext::getModule(const std::string_view name) const noexcept
{
    if (moduleExists(name)) {
        const int r = modules.at(name);
        lua_rawgeti(L, LUA_REGISTRYINDEX, r);
        return ModuleDeclaration::CHECK(L, -1);
    }

    else {
        const std::string moduleName(name);
        luaL_error(L, "Module %s not found", moduleName.c_str());
        return nullptr;
    }
}

FunctionDeclaration *
LuaContext::getFunction(const std::string_view module, const std::string_view name) const noexcept
{
    if (moduleExists(module)) {
        const int r = functions.at({ module, name });
        lua_rawgeti(L, LUA_REGISTRYINDEX, r);
        return FunctionDeclaration::CHECK(L, -1);
    }

    else {
        const std::string moduleName(module);
        const std::string functionName(name);
        luaL_error(L, "Function %s::%s not found", moduleName.c_str(), functionName.c_str());
        return nullptr;
    }
}

JobDeclaration *
LuaContext::getJob(const std::string_view module, const std::string_view name) const noexcept
{
    if (moduleExists(module) && jobs.count({ module, name }) >= 1) {
        const int r = jobs.at({ module, name });
        lua_rawgeti(L, LUA_REGISTRYINDEX, r);
        return JobDeclaration::CHECK(L, -1);
    }

    else {
        const std::string moduleName(module);
        const std::string jobName(name);
        luaL_error(L, "Job %s::%s not found", moduleName.c_str(), jobName.c_str());
        return nullptr; // Unreachable
    }
}

const std::vector<std::string_view>
LuaContext::getAllModuleName() const noexcept
{
    std::vector<std::string_view> moduleNames{};
    for (auto const &[name, _] : modules)
        moduleNames.emplace_back(name);
    return moduleNames;
}

const std::string &
LuaContext::getCurrentLuaFile() const
{
    if (currentFile.empty())
        throw std::logic_error("Not called when a file is being evaluated");
    return currentFile;
}
