#pragma once

#include "PreCompiledHeaders.hh"
#include "VivyApplication.hh"
#include "Lib/Log.hh"
#include "Lib/Utils.hh"

struct lua_State;

namespace Vivy
{
class ScriptDocument;
class VivyDocument;
}

namespace Vivy::Script
{
class ScriptStore;
class ModuleDeclaration;
class FunctionDeclaration;
class JobDeclaration;
}

namespace Vivy::Script
{
// New Lua script instance
class LuaContext final {
    VIVY_UNMOVABLE_OBJECT(LuaContext)
    VIVY_APP_LOGGABLE_OBJECT(LuaContext, logger)

    lua_State *L{ nullptr };
    std::shared_ptr<Vivy::VivyDocument> attachedVivyDocument{ nullptr };

    std::string failureString{};
    std::string currentFile{};
    std::string lastLoadedModule{};

    std::map<const std::string_view, const int> modules                                       = {};
    std::map<std::tuple<const std::string_view, const std::string_view>, const int> functions = {};
    std::map<std::tuple<const std::string_view, const std::string_view>, const int> jobs      = {};

    static inline std::map<const lua_State *const, LuaContext *const> contextList = {};

public:
    enum class Code { Success, Error };
    static constexpr inline Code Success = Code::Success;
    static constexpr inline Code Error   = Code::Error;

    LuaContext() noexcept;
    ~LuaContext() noexcept;

    Code loadDocument(std::shared_ptr<ScriptDocument>) noexcept;
    Code loadDocument(std::shared_ptr<Vivy::VivyDocument>) noexcept;

    QString getLastLuaError() noexcept;

    // Fail and abort the LuaContext
    [[noreturn]] void failedWith(const std::string &) noexcept;
    [[noreturn]] void failedWith(const char *) noexcept;

    // Set the error message for the lua context and continue
    void setFailedWith(const char *) noexcept;
    void setFailedWith(const std::string &) noexcept;

    lua_State *getState() noexcept;
    std::shared_ptr<Vivy::VivyDocument> getAttachedVivyDocument() const noexcept;

    decltype(auto) getOutputStream() const noexcept { return logInfo(); }
    decltype(auto) getErrorStream() const noexcept { return logError(); }

    operator lua_State *() noexcept { return L; }
    static LuaContext *getContext(const lua_State *const) noexcept;

    // WARN: Need the ModuleDeclaration, the FunctionDeclaration and the
    // JobDeclaration to be on top of the stack!
    Code registerDeclaration(const ModuleDeclaration *const) noexcept;
    Code registerDeclaration(const FunctionDeclaration *const) noexcept;
    Code registerDeclaration(const JobDeclaration *const) noexcept;

    // Returns the module or `nullptr` if not found.
    ModuleDeclaration *getModule(const std::string_view) const noexcept;
    FunctionDeclaration *getFunction(const std::string_view, const std::string_view) const noexcept;
    JobDeclaration *getJob(const std::string_view, const std::string_view) const noexcept;
    const std::vector<std::string_view> getAllModuleName() const noexcept;

    // Makes only sens when called from within loadFile and loadPackagedFile
    const std::string &getCurrentLuaFile() const;
    void setLastLoadedModule(const std::string &) noexcept;

private:
    // Exec all loaded strings and files
    Code exec() noexcept;

    Code loadString(const char *) noexcept;
    Code loadFile(const char *) noexcept;
    void loadPackagedFile(const QString &) noexcept;

    Code registerToMapTuple(const std::string_view module, const std::string_view name,
                            std::map<std::tuple<const std::string_view, const std::string_view>,
                                     const int> *const mapPtr) noexcept;
    bool moduleExists(const std::string_view) const noexcept;

    // Swap the env, needed before executing a user script!
    void swapEnv() noexcept;
};

[[maybe_unused]] static inline decltype(auto)
out(const LuaContext *const context) noexcept
{
    return context->getOutputStream();
}

[[maybe_unused]] static inline decltype(auto)
err(const LuaContext *const context) noexcept
{
    return context->getErrorStream();
}

}
