#pragma once

#ifndef __cplusplus
#error "This is a C++ header"
#endif

#include "PreCompiledHeaders.hh"
#include "VivyApplication.hh"
#include "Lib/Utils.hh"
#include "Lib/Log.hh"
#include "Lib/AbstractDocument.hh"
#include "Lib/Document/VivyDocumentStore.hh"
#include "UI/DocumentViews/AudioVisualizer.hh"
#include "UI/VivyDocumentView.hh"
#include "UI/ScriptDocumentView.hh"

namespace Vivy
{
class AboutWindow;

class MainWindow final : public QMainWindow {
    Q_OBJECT
    VIVY_APP_LOGGABLE_OBJECT(MainWindow, logger)

    QTabWidget *documents{ nullptr };
    QMenu *viewMenu{ nullptr };

    QMutex aboutWindowMutex;
    AboutWindow *aboutWindow{ nullptr };

public:
    explicit MainWindow() noexcept;

    AbstractDocument *getCurrentDocument() const;
    template <class Document> Document *getCurrentDocument() const
    {
        if (AbstractDocumentView *currentView = getCurrentDocumentView();
            currentView->getType() == Document::type) {
            return currentView->getDocument();
        } else {
            throw std::runtime_error("No main document view was found");
        }
    }

public slots:
    void closeDocument(AbstractDocumentView *const) noexcept;

private:
    void addTab(AbstractDocumentView *const);
    AbstractDocumentView *getTab(const int) const noexcept;
    AbstractDocumentView *getCurrentDocumentView() const;

    int getIndexOfTab(const AbstractDocumentView *const) const noexcept;
    int findFirstUntouchedDocument() const noexcept;
    void updateFakeVimUsage(bool yes) noexcept;
    QString dialogOpenFileName(const QString &title, const QString &folder,
                               const QString &filter) noexcept;
    QString dialogSaveFileName(const QString &title, const QString &folder,
                               const QString &filter) noexcept;

    QString executeDialog(MainWindow *const self, QFileDialog *const dialog) noexcept;

    // Do an action with the selected filename. The 'call' variable must be
    // callable like this: call(DocumentView, Document, QString)
    template <typename DV, typename D>
    void withOpenFileNameDialog(const QString &title, const QString &filter, auto call) noexcept
    {
        const QString filename = dialogOpenFileName(title, QDir::homePath(), filter);

        if (filename.isEmpty()) {
            logWarning() << "Found an empty filename, don't open a file";
            return;
        }

        try {
            if (AbstractDocumentView *currView = getCurrentDocumentView();
                currView->isType(D::type)) {
                DV *view = static_cast<VivyDocumentView *>(currView);
                D *doc   = view->getDocument();
                call(view, doc, filename);
            }
        } catch (const std::runtime_error &e) {
            logError() << "Operation failed with: " << e.what();
        }
    }

    // Walk throu all views in the QTabWidget and also get the index of the tab
    template <Derived<AbstractDocumentView> DocViewType> void forEachViews(auto callback) const
    {
        const int count = documents->count();
        for (int index = 0; index < count; ++index) {
            AbstractDocumentView *docView = getTab(index);
            if constexpr (std::is_same_v<AbstractDocumentView, DocViewType>)
                callback(docView, index);
            else if (DocViewType *view = dynamic_cast<DocViewType *>(docView); view != nullptr)
                callback(view, index);
        }
    }

protected:
    void closeEvent(QCloseEvent *) noexcept override;

private slots:
    void newDocument() noexcept;
    void openDocument() noexcept;
    void closeDocument(int index) noexcept;

    void openProperties(int index) noexcept;

    void loadSubDocumentAss() noexcept;
    void loadSubDocumentVideo() noexcept;
    void loadSubDocumentAudio() noexcept;

    void saveFile() noexcept;
    void saveFileAs() noexcept;
    void renameFile() noexcept;

    void openDialogHelp() noexcept;

    void documentViewActionsChanged() noexcept;
};
}
