#include "PreCompiledHeaders.hh"
#include "UI/PropertyModel.hh"

using namespace Vivy;

PropertyModel::Item::Item(Item *theParent) noexcept
    : parent(theParent)
{
}

PropertyModel::Item::~Item() noexcept { qDeleteAll(childs); }

void
PropertyModel::Item::appendChild(PropertyModel::Item *item) noexcept
{
    childs.append(item);
}

PropertyModel::Item *
PropertyModel::Item::getChild(int row) const noexcept
{
    return childs.value(row);
}

PropertyModel::Item *
PropertyModel::Item::getParent() const noexcept
{
    return parent;
}

int
PropertyModel::Item::getChildCount() const noexcept
{
    return childs.count();
}

// Can't be const because of indexOf...
int
PropertyModel::Item::getRow() noexcept
{
    return parent ? parent->childs.indexOf(this) : 0;
}

void
PropertyModel::Item::setKey(const QString &k) noexcept
{
    key = k;
}

void
PropertyModel::Item::setValue(const QString &v) noexcept
{
    value = v;
}

void
PropertyModel::Item::setType(const QJsonValue::Type &t) noexcept
{
    type = t;
}

QString
PropertyModel::Item::getKey() const noexcept
{
    return key;
}

QString
PropertyModel::Item::getValue() const noexcept
{
    return value;
}

QJsonValue::Type
PropertyModel::Item::getType() const noexcept
{
    return type;
}

const QVector<PropertyModel::Item *>
PropertyModel::Item::getChilds() const noexcept
{
    return childs;
}

bool
PropertyModel::Item::isOrderedJsonObject(const QJsonValue &value) noexcept
{
    if (value.isArray()) {
        for (const QJsonValue v : value.toArray()) {
            if (!v.isObject() || v.toObject().count() != 1)
                return false;
        }
        return true;
    }
    return false;
}

PropertyModel::Item *
PropertyModel::Item::createChild(Item *root, QJsonObject::ConstIterator const &it) noexcept
{
    const QString key  = it.key();
    const QJsonValue v = it.value();
    Item *child        = fromJson(v, root);
    child->setKey(key);
    child->setType(v.type());
    return child;
}

PropertyModel::Item *
PropertyModel::Item::createChild(Item *root, QJsonObject const &singleElem) noexcept
{
    Q_ASSERT(singleElem.count() == 1);

    const QString key    = singleElem.keys().at(0);
    const QJsonValue val = singleElem.value(key);

    Item *child = fromJson(val, root);
    child->setKey(key);
    child->setType(val.type());

    return child;
}

PropertyModel::Item *
PropertyModel::Item::createChild(Item *root, int pos, QJsonValue const &val) noexcept
{
    Item *child = fromJson(val, root);
    child->setKey(QString::number(pos));
    child->setType(val.type());
    return child;
}

PropertyModel::Item *
PropertyModel::Item::fromJson(const QJsonValue &value, PropertyModel::Item *parent)
{
    Item *root = new Item(parent);
    root->setKey("root");

    if (value.isObject()) {
        const QJsonObject objValue = value.toObject();
        const auto end             = objValue.constEnd();
        for (auto it = objValue.constBegin(); it != end; ++it)
            root->appendChild(createChild(root, it));
    }

    else if (isOrderedJsonObject(value)) {
        for (const QJsonValue v : value.toArray())
            root->appendChild(createChild(root, v.toObject()));
    }

    else if (value.isArray()) {
        for (int index = 0; const QJsonValue v : value.toArray()) {
            root->appendChild(createChild(root, index, v));
            ++index;
        }
    }

    else {
        root->setValue(value.toVariant().toString());
        root->setType(value.type());
    }

    return root;
}

PropertyModel::PropertyModel(QObject *parent) noexcept
    : QAbstractItemModel(parent)
{
    headers.append(QStringLiteral("Key"));
    headers.append(QStringLiteral("Value"));
}

PropertyModel::PropertyModel(const QJsonDocument &properties, QObject *parent)
    : PropertyModel(parent)
{
    loadJson(properties);
}

void
PropertyModel::loadJson(const QJsonDocument &jdoc)
{
    if (!jdoc.isNull()) {
        beginResetModel();
        root.reset();

        if (jdoc.isArray()) {
            root.reset(Item::fromJson(QJsonValue(jdoc.array())));
            root->setType(QJsonValue::Array);

        } else {
            root.reset(Item::fromJson(QJsonValue(jdoc.object())));
            root->setType(QJsonValue::Object);
        }

        endResetModel();
        return;
    }

    throw std::runtime_error("failed to load json");
}

QVariant
PropertyModel::data(const QModelIndex &index, int role) const noexcept
{
    if (!index.isValid())
        return QVariant();

    const Item *item = static_cast<Item *>(index.internalPointer());

    if (role == Qt::DisplayRole) {
        if (index.column() == 0)
            return QString("%1").arg(item->getKey());

        if (index.column() == 1)
            return QString("%1").arg(item->getValue());
    }

    return QVariant();
}

bool
PropertyModel::setData(const QModelIndex &, const QVariant &, int) noexcept
{
    return false;
}

QVariant
PropertyModel::headerData(int section, Qt::Orientation orientation, int role) const noexcept
{
    if (role != Qt::DisplayRole)
        return QVariant();

    if (orientation == Qt::Horizontal)
        return headers.value(section);

    else
        return QVariant();
}

QModelIndex
PropertyModel::index(int row, int column, const QModelIndex &parent) const noexcept
{
    if (!hasIndex(row, column, parent))
        return QModelIndex();

    const Item *parentItem =
        parent.isValid() ? static_cast<Item *>(parent.internalPointer()) : root.get();
    Item *childItem = parentItem->getChild(row);
    return childItem ? createIndex(row, column, childItem) : QModelIndex();
}

QModelIndex
PropertyModel::parent(const QModelIndex &index) const noexcept
{
    if (!index.isValid())
        return QModelIndex();

    const Item *childItem = static_cast<Item *>(index.internalPointer());
    Item *parentItem      = childItem->getParent();

    if (parentItem == root.get())
        return QModelIndex();

    return createIndex(parentItem->getRow(), 0, parentItem);
}

int
PropertyModel::rowCount(const QModelIndex &parent) const noexcept
{
    if (parent.column() > 0)
        return 0;

    const Item *parentItem =
        parent.isValid() ? static_cast<Item *>(parent.internalPointer()) : root.get();
    return parentItem->getChildCount();
}

int
PropertyModel::columnCount(const QModelIndex &) const noexcept
{
    return 2;
}

Qt::ItemFlags
PropertyModel::flags(const QModelIndex &index) const noexcept
{
    return QAbstractItemModel::flags(index);
}

QTreeView *
PropertyModel::getView(QWidget *parent, PropertyModel *ptr) noexcept
{
    QTreeView *view = new QTreeView(parent);
    view->setModel(ptr);
    view->header()->setSectionResizeMode(QHeaderView::ResizeToContents);
    view->expandAll();
    return view;
}
