#include "PreCompiledHeaders.hh"
#include "UI/ScriptDocumentView.hh"
#include "UI/MainWindow.hh"
#include "UI/ScriptViews/ScriptEditor.hh"
#include "UI/ScriptViews/ScriptHighlighter.hh"
#include "VivyApplication.hh"

using namespace Vivy;

ScriptDocumentView::ScriptDocumentView(std::shared_ptr<ScriptDocument> ptr, QWidget *parent)
    : AbstractDocumentView(AbstractDocumentView::Type::Script, parent)
    , editor(new ScriptEditor(this))
    , syntax(new ScriptHighlighter(editor->document()))
    , document(ptr)
{
    QFile textFile(ptr->getName());
    if (!textFile.open(QIODevice::ReadOnly | QIODevice::Text)) {
        throw std::runtime_error("Failed to open script file");
    }

    editor->setPlainText(QString::fromUtf8(textFile.readAll()));
    setCentralWidget(editor);
    editor->setFocus(Qt::OtherFocusReason);

    setUseFakeVimEditor(vivyApp->getUseFakeVimEditor());

    connect(this, &ScriptDocumentView::luaErrorFound, editor, &ScriptEditor::updateLastLuaError);
    connect(document.get(), &AbstractDocument::documentChanged, this,
            [=, this]() noexcept -> void { emit documentPropertyChanged(); });

    // Same style as the editor
    setStyleSheet(QStringLiteral("* {"
                                 "  background-color: #232629;"
                                 "  font-family: \"FiraCode\";"
                                 "  font-size: 10pt"
                                 "}"));

    VIVY_LOG_CTOR() << "Creating view for script " << VIVY_LOG_QUOTED(ptr->getName()) << " with "
                    << ptr->getUuid();
}

ScriptDocumentView::~ScriptDocumentView()
{
    setUseFakeVimEditor(false);
    VIVY_LOG_DTOR() << "Closing a script view";
}

void
ScriptDocumentView::setUseFakeVimEditor(bool yes) noexcept
{
    if (yes && !isUsingFakeVim) {
        MainWindow *const mw = vivyApp->getMainWindow();
        handler              = new FakeVimHandler(editor);
        proxy                = EditorProxy::connectSignals(handler, editor);
        isUsingFakeVim       = true;
        connect(proxy, &EditorProxy::handleInput, handler, &FakeVimHandler::handleInput);
        connect(proxy, &EditorProxy::requestQuit, this, [this, mw]() noexcept -> void {
            mw->closeDocument(static_cast<AbstractDocumentView *>(this));
        });
        connect(proxy, &EditorProxy::requestSave, this, [this]() { document->save(); });
        connect(proxy, &EditorProxy::requestSaveAndQuit, this, [this, mw]() noexcept -> void {
            document->save();
            mw->closeDocument(static_cast<AbstractDocumentView *>(this));
        });
        initHandler(handler);
        clearUndoRedo(editor);
    }

    else if (!yes) {
        if (handler) {
            handler->disconnectFromEditor();
            delete handler;
        }

        if (proxy)
            delete proxy;

        clearUndoRedo(editor);
        editor->setOverwriteMode(false);
        isUsingFakeVim = false;
        proxy          = nullptr;
        handler        = nullptr;
    }
}

void
ScriptDocumentView::initHandler(FakeVimHandler *handler) noexcept
{
    handler->handleCommand(QStringLiteral("set nopasskeys"));
    handler->handleCommand(QStringLiteral("set nopasscontrolkey"));
    handler->installEventFilter();
    handler->setupWidget();

    handler->handleCommand(QStringLiteral("set expandtab"));
    handler->handleCommand(QStringLiteral("set shiftwidth=4"));
    handler->handleCommand(QStringLiteral("set tabstop=4"));
    handler->handleCommand(QStringLiteral("set autoindent"));
    handler->handleCommand(QStringLiteral("set smartindent"));
}

void
ScriptDocumentView::clearUndoRedo(QPlainTextEdit *scriptEditor) noexcept
{
    scriptEditor->setUndoRedoEnabled(false);
    scriptEditor->setUndoRedoEnabled(true);
}

void
ScriptDocumentView::closeDocument() noexcept
{
}

QString
ScriptDocumentView::getDocumentTabName() const noexcept
{
    const QFileInfo fileInfo(document->getName());
    return fileInfo.completeBaseName();
}

QString
ScriptDocumentView::getDocumentTabToolTip() const noexcept
{
    return QStringLiteral("UUID: ") + document->getUuid().toString() +
           QStringLiteral("\nType: Vivy Script");
}

QIcon
ScriptDocumentView::getDocumentTabIcon() const noexcept
{
    return QIcon(VIVY_ICON_SCRIPT);
}

AbstractDocument *
ScriptDocumentView::getDocument() const noexcept
{
    return document.get();
}

void
ScriptDocumentView::openProperties() noexcept
{
}
