#include "PreCompiledHeaders.hh"
#include "UI/ScriptViews/EditorProxy.hh"
#include "UI/MainWindow.hh"
#include "UI/FakeVim/FakeVimHandler.hh"
#include "UI/FakeVim/FakeVimActions.hh"
#include "VivyApplication.hh"

using namespace Vivy;

Vivy::EditorProxy *
EditorProxy::connectSignals(FakeVimHandler *handler, QPlainTextEdit *editor) noexcept
{
    EditorProxy *proxy = new EditorProxy(editor);

    handler->commandBufferChanged.connect(
        [proxy](const QString &contents, int cursorPos, int, int) noexcept -> void {
            proxy->changeStatusMessage(contents, cursorPos);
        });

    handler->extraInformationChanged.connect(
        [proxy](const QString &text) noexcept -> void { proxy->changeExtraInformation(text); });
    handler->statusDataChanged.connect(
        [proxy](const QString &text) noexcept -> void { proxy->changeStatusData(text); });
    handler->highlightMatches.connect(
        [proxy](const QString &needle) noexcept -> void { proxy->highlightMatches(needle); });
    handler->handleExCommandRequested.connect(
        [proxy](bool *handled, const ExCommand &cmd) noexcept -> void {
            proxy->handleExCommand(handled, cmd);
        });
    handler->requestSetBlockSelection.connect([proxy](const QTextCursor &cursor) noexcept -> void {
        proxy->requestSetBlockSelection(cursor);
    });
    handler->requestDisableBlockSelection.connect(
        [proxy]() noexcept -> void { proxy->requestDisableBlockSelection(); });
    handler->requestHasBlockSelection.connect(
        [proxy](bool *on) noexcept -> void { proxy->requestHasBlockSelection(on); });

    handler->indentRegion.connect(
        [proxy](int beginBlock, int endBlock, QChar typedChar) noexcept -> void {
            proxy->indentRegion(beginBlock, endBlock, typedChar);
        });
    handler->checkForElectricCharacter.connect([proxy](bool *result, QChar c) noexcept -> void {
        proxy->checkForElectricCharacter(result, c);
    });

    return proxy;
}

EditorProxy::EditorProxy(QPlainTextEdit *widg) noexcept
    : QObject()
    , widget(widg)
{
    VIVY_LOG_CTOR()
        << "Creating new proxy for a the text view " << VIVY_LOG_QUOTED(widg->windowFilePath());
}

EditorProxy::~EditorProxy() { VIVY_LOG_DTOR() << "Delete editor proxy!"; }

void
EditorProxy::changeStatusData(const QString &info) noexcept
{
    statusData = info;
    updateStatusBar();
}

void
EditorProxy::highlightMatches(const QString &pattern) noexcept
{
    QTextDocument *doc = widget->document();
    Q_ASSERT(doc);

    QTextEdit::ExtraSelection selection;
    selection.format.setBackground(Qt::yellow);
    selection.format.setForeground(Qt::black);

    // Highlight matches.
    QRegExp re(pattern);
    QTextCursor cur = doc->find(re);
    searchSelection.clear();

    int a = cur.position();
    while (!cur.isNull()) {
        if (cur.hasSelection()) {
            selection.cursor = cur;
            searchSelection.append(selection);
        } else {
            cur.movePosition(QTextCursor::NextCharacter);
        }

        cur   = doc->find(re, cur);
        int b = cur.position();

        if (a == b) {
            cur.movePosition(QTextCursor::NextCharacter);
            cur = doc->find(re, cur);
            b   = cur.position();
            if (a == b)
                break;
        }

        a = b;
    }

    updateExtraSelections();
}

void
EditorProxy::changeStatusMessage(const QString &contents, int cursorPos) noexcept
{
    statusMessage = (cursorPos == -1)
                        ? contents
                        : contents.left(cursorPos) + QChar(10073) + contents.mid(cursorPos);
    updateStatusBar();
}

void
EditorProxy::changeExtraInformation(const QString &info) noexcept
{
    QMessageBox::information(widget, tr("Information"), info);
}

void
EditorProxy::updateStatusBar() noexcept
{
    vivyApp->getMainWindow()->statusBar()->showMessage(statusMessage); // + statusData
}

void
EditorProxy::handleExCommand(bool *handled, const ExCommand &cmd) noexcept
{
    // :wq
    if (wantSaveAndQuit(cmd))
        emit requestSaveAndQuit();

    // :w
    else if (wantSave(cmd))
        emit requestSave();

    else if (wantQuit(cmd)) {
        // :q!
        if (cmd.hasBang) {
            logInfo() << "Want to force quit the editor!";
            emit requestQuit();
        }

        // :q
        else {
            logInfo() << "Want to quit the editor";
            emit requestQuit();
        }
    }

    else if (wantRun(cmd))
        emit requestRun();

    else {
        *handled = false;
        return;
    }

    *handled = true;
}

void
EditorProxy::requestSetBlockSelection(const QTextCursor &tc) noexcept
{
    const QPalette pal = widget->parentWidget() != nullptr ? widget->parentWidget()->palette()
                                                           : QApplication::palette();

    blockSelection.clear();
    clearSelection.clear();

    QTextCursor cur = tc;

    QTextEdit::ExtraSelection selection;
    selection.format.setBackground(pal.color(QPalette::Base));
    selection.format.setForeground(pal.color(QPalette::Text));
    selection.cursor = cur;
    clearSelection.append(selection);

    selection.format.setBackground(pal.color(QPalette::Highlight));
    selection.format.setForeground(pal.color(QPalette::HighlightedText));

    const int from = cur.positionInBlock();
    const int to   = cur.anchor() - cur.document()->findBlock(cur.anchor()).position();
    const int min  = qMin(cur.position(), cur.anchor());
    const int max  = qMax(cur.position(), cur.anchor());
    for (QTextBlock block                                 = cur.document()->findBlock(min);
         block.isValid() && block.position() < max; block = block.next()) {
        cur.setPosition(block.position() + qMin(from, block.length()));
        cur.setPosition(block.position() + qMin(to, block.length()), QTextCursor::KeepAnchor);
        selection.cursor = cur;
        blockSelection.append(selection);
    }

    disconnect(widget, &QPlainTextEdit::selectionChanged, this, &EditorProxy::updateBlockSelection);
    widget->setTextCursor(tc);
    connect(widget, &QPlainTextEdit::selectionChanged, this, &EditorProxy::updateBlockSelection);

    QPalette pal2 = widget->palette();
    pal2.setColor(QPalette::Highlight, Qt::transparent);
    pal2.setColor(QPalette::HighlightedText, Qt::transparent);
    widget->setPalette(pal2);

    updateExtraSelections();
}

void
EditorProxy::requestDisableBlockSelection() noexcept
{
    const QPalette pal = widget->parentWidget() != nullptr ? widget->parentWidget()->palette()
                                                           : QApplication::palette();

    blockSelection.clear();
    clearSelection.clear();
    widget->setPalette(pal);

    disconnect(widget, &QPlainTextEdit::selectionChanged, this, &EditorProxy::updateBlockSelection);

    updateExtraSelections();
}

void
EditorProxy::updateBlockSelection() noexcept
{
    requestSetBlockSelection(widget->textCursor());
}

void
EditorProxy::requestHasBlockSelection(bool *on) noexcept
{
    *on = !blockSelection.isEmpty();
}

void
EditorProxy::indentRegion(int beginBlock, int endBlock, QChar typedChar) noexcept
{
    QTextDocument *doc = widget->document();
    Q_ASSERT(doc);

    const int indentSize  = static_cast<int>(fakeVimSettings()->shiftWidth.value());
    QTextBlock startBlock = doc->findBlockByNumber(beginBlock);

    // Record line lenghts for mark adjustments
    QVector<int> lineLengths(endBlock - beginBlock + 1);
    QTextBlock block = startBlock;

    for (int i = beginBlock; i <= endBlock; ++i) {
        const QString line          = block.text();
        lineLengths[i - beginBlock] = line.length();

        if (typedChar.unicode() == 0 && line.simplified().isEmpty()) {
            // clear empty lines
            QTextCursor cursor(block);
            while (!cursor.atBlockEnd())
                cursor.deleteChar();
        }

        else {
            const QTextBlock previousBlock = block.previous();
            const QString previousLine = previousBlock.isValid() ? previousBlock.text() : QString();

            int indent = firstNonSpace(previousLine);
            if (typedChar == '}')
                indent = std::max(0, indent - indentSize);
            else if (previousLine.endsWith("{"))
                indent += indentSize;
            const QString indentString = QString(" ").repeated(indent);

            QTextCursor cursor(block);
            cursor.beginEditBlock();
            cursor.movePosition(QTextCursor::StartOfBlock);
            cursor.movePosition(QTextCursor::NextCharacter, QTextCursor::KeepAnchor,
                                firstNonSpace(line));
            cursor.removeSelectedText();
            cursor.insertText(indentString);
            cursor.endEditBlock();
        }

        block = block.next();
    }
}

void
EditorProxy::checkForElectricCharacter(bool *result, QChar c) noexcept
{
    *result = c == '{' || c == '}';
}

int
EditorProxy::firstNonSpace(const QString &text) noexcept
{
    int indent = 0;
    while (indent < text.length() && text.at(indent) == ' ')
        ++indent;
    return indent;
}

void
EditorProxy::updateExtraSelections() noexcept
{
    widget->setExtraSelections(clearSelection + searchSelection + blockSelection);
}

bool
EditorProxy::wantSaveAndQuit(const ExCommand &cmd) noexcept
{
    return cmd.cmd == "wq";
}

bool
EditorProxy::wantSave(const ExCommand &cmd) noexcept
{
    logInfo() << "Want to save the editor's file";
    return cmd.matches("w", "write") || cmd.matches("wa", "wall");
}

bool
EditorProxy::wantQuit(const ExCommand &cmd) noexcept
{
    logInfo() << "Want to quit editor";
    return cmd.matches("q", "quit") || cmd.matches("qa", "qall");
}

bool
EditorProxy::wantRun(const ExCommand &cmd) noexcept
{
    logInfo() << "Want to run a command: " << cmd.cmd;
    return cmd.matches("run", "run") || cmd.matches("make", "make");
}
