#include "PreCompiledHeaders.hh"
#include "UI/VivyDocumentView.hh"
#include "UI/PropertyModel.hh"
#include "UI/Utils.hh"
#include "UI/DocumentViews/AudioVisualizer.hh"
#include "UI/DocumentViews/AssLinesView.hh"
#include "UI/DocumentViews/AssLinesModel.hh"
#include "VivyApplication.hh"
#include "Lib/Document/VivyDocument.hh"
#include "Lib/Document/VivyDocumentStore.hh"

using namespace Vivy;

VivyDocumentView::VivyDocumentView(std::shared_ptr<VivyDocument> doc, QWidget *parent) noexcept
    : AbstractDocumentView(AbstractDocumentView::Type::Vivy, parent)
    , document(doc)
{
    VIVY_LOG_CTOR() << "Creating view for vivy document " << VIVY_LOG_QUOTED(doc->getName())
                    << " with " << doc->getUuid() << "...";

    setDockNestingEnabled(true);

    loadAudioView();
    loadVideoView();
    loadAssView();

    // Add some actions...
    openPropertiesAct = new QAction("Open properties", this);
    connect(openPropertiesAct, &QAction::triggered, this, &VivyDocumentView::openProperties);
    connect(document.get(), &AbstractDocument::documentChanged, this, [=, this]() {
        if (property)
            openProperties();
        emit documentPropertyChanged();
    });
    viewsActions.append(openPropertiesAct);

    // The separator
    QAction *const separator = new QAction(this);
    separator->setSeparator(true);
    viewsActions.append(separator);

    VIVY_LOG_CTOR() << "View for " << doc->getUuid() << " done!";
}

VivyDocumentView::~VivyDocumentView() noexcept
{
    VIVY_LOG_DTOR() << "Deleting view for " << document->getUuid() << ", file name is "
                    << VIVY_LOG_QUOTED(document->getName());
    deleteAllContent();
}

VivyDocument *
VivyDocumentView::getDocument() const noexcept
{
    return document.get();
}

QString
VivyDocumentView::getDocumentTabName() const noexcept
{
    return (document->checkDocumentOption(VivyDocument::UntouchedByDefault)
                ? (Utils::untouchedTabIndicator + " ")
                : QString::fromUtf8("")) +
           Utils::getBaseName(document->getName());
}

QString
VivyDocumentView::getDocumentTabToolTip() const noexcept
{
    return QStringLiteral("UUID: ") + document->getUuid().toString() +
           QStringLiteral("\nType: Vivy Document {") + document->getDocumentCapabilitiesString() +
           QStringLiteral("}") +
           (document->checkDocumentOption(VivyDocument::UntouchedByDefault)
                ? QStringLiteral("\nUntouched document")
                : QStringLiteral("")) +
           (document->checkDocumentOption(VivyDocument::MemoryDocumentCreation)
                ? QStringLiteral("\nMemory document")
                : QStringLiteral(""));
}

void
VivyDocumentView::loadVideoView() noexcept
{
    if (!document->checkDocumentCapabilities(VivyDocument::Capabilities::VideoAble))
        return;

    if (!videoView) {
        videoView = new UnclosableDockWidget("Video View", this);
        videoView->setAllowedAreas(Qt::AllDockWidgetAreas);
        videoView->setFeatures(allDockFeatures);
        addDockWidget(Qt::BottomDockWidgetArea, videoView, Qt::Vertical);
    }

    Utils::deleteInternalWidget(videoView);
    videoView->setWidget(new VideoView(videoView));
    qobject_cast<VideoView *>(videoView->widget())
        ->loadFile(document->getVideoSubDocument()->getFilePath());
}

void
VivyDocumentView::loadAssView() noexcept
{
    if (!document->checkDocumentCapabilities(VivyDocument::Capabilities::AssAble))
        return;

    if (!assLines) {
        assLines = new UnclosableDockWidget("ASS Lines", this);
        assLines->setFeatures(onlyClosableDockFeatures);
        assLines->setAllowedAreas(Qt::LeftDockWidgetArea | Qt::RightDockWidgetArea |
                                  Qt::BottomDockWidgetArea);
        addDockWidget(Qt::BottomDockWidgetArea, assLines, Qt::Vertical);
    }

    assModel.reset(new AssLinesModel(document->getAssSubDocument()->getLines()));
    Utils::deleteInternalWidget(assLines);
    assLines->setWidget(new AssLinesView(assModel.get(), assLines));
}

void
VivyDocumentView::loadAudioView() noexcept
{
    if (!document->checkDocumentCapabilities(VivyDocument::AudioAble))
        return;

    std::shared_ptr<AudioSubDocument> audioDocument = document->getAudioSubDocument();
    AudioContext::StreamPtr stream                  = audioDocument->getDefaultStream();

    if (stream == nullptr) {
        logError() << "Failed to get default audio stream";
        return;
    }

    if (!visualizer) {
        visualizer = new UnclosableDockWidget("Visualizer", this);
        visualizer->setSizePolicy(QSizePolicy::Ignored, QSizePolicy::Minimum);
        visualizer->setAllowedAreas(Qt::LeftDockWidgetArea | Qt::TopDockWidgetArea |
                                    Qt::BottomDockWidgetArea);
        visualizer->setFeatures(onlyClosableDockFeatures);
        addDockWidget(Qt::LeftDockWidgetArea, visualizer, Qt::Horizontal);
    }

    // Kubat: don't check, may throw an error but don't think we can
    // recover from it.
    AudioVisualizer *visualizerInner = new AudioVisualizer(stream, visualizer);
    Utils::deleteInternalWidget(visualizer);
    visualizer->setWidget(visualizerInner);
    visualizer->layout()->setAlignment(visualizerInner, Qt::AlignTop);
}

void
VivyDocumentView::closeDocument() noexcept
{
    logDebug() << "Closing document " << document->getUuid();
    vivyApp->documentStore->closeDocument(document->getUuid());
    allowToCloseAllDocks();
}

QIcon
VivyDocumentView::getDocumentTabIcon() const noexcept
{
    return QIcon(VIVY_ICON_APP);
}

void
VivyDocumentView::openProperties() noexcept
{
    propertyModel   = std::make_unique<PropertyModel>(*document.get());
    QTreeView *view = PropertyModel::getView(property, propertyModel.get());

    if (!property) {
        property = new UnclosableDockWidget("Properties", this);
        property->setAllowedAreas(Qt::AllDockWidgetAreas);
        property->setFeatures(allDockFeatures);
        addDockWidget(Qt::RightDockWidgetArea, property, Qt::Vertical);
    }

    property->setWidget(view);
    openPropertiesAct->setVisible(false);
}
