#pragma once

#ifndef __cplusplus
#error "This is a C++ header"
#endif

#include "PreCompiledHeaders.hh"

// Get the VivyApplication pointer
#define vivyApp (static_cast<::Vivy::VivyApplication *>(VivyApplication::instance()))

// Install logger, use the global LogSink
#define VIVY_APP_LOGGABLE_OBJECT(name, logger) \
    VIVY_LOGGABLE_OBJECT(vivyApp->getLogSink(), name, logger)

// Install logger, use the global LogSink, with a stored name like in the
// VIVY_LOGGABLE_OBJECT_BY_STORED_NAME macro.
#define VIVY_APP_LOGGABLE_OBJECT_BY_STORED_NAME(name, logger) \
    VIVY_LOGGABLE_OBJECT_BY_STORED_NAME(vivyApp->getLogSink(), name, logger)

#define currentVivyDocument   dynamic_cast<VivyDocument *>(vivyApp->getCurrentDocument())
#define currentScriptDocument dynamic_cast<ScriptDocument *>(vivyApp->getCurrentDocument())

// Only support dark theme for now
#define VIVY_ICON_APP     ":icons/vivy.png"
#define VIVY_ICON_SCRIPT  ":icons/lua.png"
#define VIVY_ICON_OPEN    ":icons/dark/document-open.svg"
#define VIVY_ICON_NEW     ":icons/dark/document-new.svg"
#define VIVY_ICON_SAVE    ":icons/dark/document-save.svg"
#define VIVY_ICON_SAVE_AS ":icons/dark/document-save-as.svg"
#define VIVY_ICON_RENAME  ":icons/dark/document-rename.svg"
#define VIVY_ICON_ABOUT   ":icons/dark/help-about.svg"
#define VIVY_ICON_FILE    ":icons/dark/text-x-generic.svg"
#define VIVY_ICON_FOLDER  ":icons/dark/folder.svg"
#define VIVY_ICON_PLAY    ":icons/dark/media-play.svg"
#define VIVY_ICON_PAUSE   ":icons/dark/media-pause.svg"
#define VIVY_ICON_STOP    ":icons/dark/media-stop.svg"
#define VIVY_ICON_RUN     ":icons/dark/run.svg"

// Detect MacOS
#if defined(Q_OS_DARWIN) || defined(Q_OS_MACOS)
#define VIVY_MACOS 1
#else
#define VIVY_MACOS 0
#endif

#include "Lib/Utils.hh"
#include "Lib/Log.hh"
#include "Lib/HostOsInfo.hh"

namespace Vivy
{
class MainWindow;
class ScriptStore;
class VivyDocumentStore;
class AbstractDocument;
class ScriptDocument;
class VivyDocument;
class Theme;

// Vivy application class
class VivyApplication : public QApplication {
    Q_OBJECT
    VIVY_DCL_LOG_SINK(logSink)
    VIVY_DCL_LOG_DISPATCH(logSink, stderrLogDispathcer, StderrLogSinkDispatcher)
    VIVY_LOGGABLE_OBJECT(logSink, APPLICATION, logger)

    enum class ApplicationType { GUI, CLI };

public:
    std::shared_ptr<VivyDocumentStore> documentStore;
    std::shared_ptr<ScriptStore> scriptStore;

    enum class Font {
        Monospace,
        MonospaceBold,
        Default,
        DefaultItalic,
        DefaultBold,
        DefaultBoldItalic
    };

private:
    int fontIdMonospace;
    int fontIdMonospaceBold;
    int fontIdRegular;
    int fontIdItalic;
    int fontIdBold;
    int fontIdBoldItalic;

    std::unique_ptr<Vivy::Theme> qtCreatorThemePtr{ nullptr };
    std::unique_ptr<MainWindow> mainWindowPtr{ nullptr };
    bool useFakeVim{ false };
    ApplicationType selectedType{ ApplicationType::GUI };

    // ApplicationType::CLI Only!
    std::shared_ptr<ScriptDocument> selectedScriptDoc{ nullptr };
    std::shared_ptr<VivyDocument> selectedVivyDoc{ nullptr };

public:
    VivyApplication(int &argc, char **argv);
    ~VivyApplication() noexcept override;

    int exec() noexcept;

    [[nodiscard]] QFont getApplicationFont(Font) const noexcept;
    [[nodiscard]] MainWindow *getMainWindow() const;
    [[nodiscard]] AbstractDocument *getCurrentDocument() const;
    [[nodiscard]] bool getUseFakeVimEditor() const noexcept;
    [[nodiscard]] const Theme *getTheme() const noexcept { return qtCreatorThemePtr.get(); }
    [[nodiscard]] std::shared_ptr<LogSink> getLogSink() noexcept { return logSink; }

    void setUseFakeVimEditor(bool) noexcept;
    void setTheme(ThemeType, VivyThemeType) noexcept;
};

}
